#ifndef __DATAPOINT_H__
#define __DATAPOINT_H__

#include "../std.h"
#include <string>
#include <list>
#include "../transfer/url_map.h"
#include "../misc/globus_modules.h"

class RCManager;

/*
  Abstraction of url. Also talks to meta-data databases (RC, RLS).
*/

class DataPoint {
  friend ostream& operator<<(ostream& o,const DataPoint &point);
 public:
  class Location {
   public:
    string meta;
    string url;
    Location(void) { };
    Location(const char* url_):meta(""),url(url_) { };
    Location(const char* meta_,const char* url_):meta(meta_),url(url_?url_:"") { };
    Location(const string &url_):meta(""),url(url_) { };
    Location(const string &meta_,const string &url_):meta(meta_),url(url_) { };
  };
  list<Location> locations;
 private:
  bool is_valid;
  string url;
  string common_url_options;
  list<Location>::iterator location;
  bool is_secure;
  bool is_resolved;
/*  bool globus_module_active; */

  bool is_metaexisting;

  unsigned long long int meta_size_;
  bool meta_size_valid;
  unsigned long long int meta_checksum_;
  bool meta_checksum_valid;
  time_t meta_created_;
  bool meta_created_valid;
  time_t meta_validtill_;
  bool meta_validtill_valid;
/*  bool meta_valid; */

  /* TODO - make it union */
  /* file */
  bool is_file;
  /* ftp */
  bool is_ftp;
  /* http */
  bool is_http;
  /* RC */
  string rc_url;
  RCManager *rc_mgr;
  string rc_lfn;
  bool is_rc;
  /* rls */
  string rls_url;
  string rls_lfn;
  bool is_rls;
  /* bbftp */
  bool is_bbftp;
  /* magda */
  bool is_magda;
  /*  retries  */
  int tries_left;
  /* globus modules */
#ifdef HAVE_RLS_CLIENT
  GlobusModuleCommon common_mod;
#endif
  GlobusModuleError error_mod;
  GlobusModuleGSIGSSAPI gsi_gssapi_mod;
  GlobusModuleReplicaCatalog replica_mod;
#ifdef HAVE_RLS_CLIENT
  GlobusModuleIO io_mod;
  GlobusModuleRLSClient rls_mod;
#endif

  bool meta_resolve_rc(bool source);
  bool meta_preregister_rc(bool replication);
  bool meta_postregister_rc(bool replication,bool failure);
  bool meta_preunregister_rc(bool replication);
  bool meta_unregister_rc(bool all);
  bool process_meta_url_rc(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations);
  bool meta_resolve_rls(bool source);
  bool meta_preregister_rls(bool replication);
  bool meta_postregister_rls(bool replication,bool failure);
  bool meta_preunregister_rls(bool replication);
  bool meta_unregister_rls(bool all);
  bool process_meta_url_rls(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations);

 public:
  DataPoint(const char* url);
  ~DataPoint(void);
  /*
    resolve locations and obtain metadata about file from databse
  */
  bool meta_resolve(bool source);
  bool meta_resolve(bool source,const UrlMap &maps);
  /*
    perform registration in meta database. *pre* functions should be
    called before uploading, *post* - after. *un* is used to revome
    information from database. If all=true information about logical
    file name itself is removed.
  */
  bool meta_preregister(bool replication);
  bool meta_postregister(bool replication,bool failure);
  bool meta_preunregister(bool replication);
  bool meta_unregister(bool all);

  /*
    Set and get corresponding meta-information related to lfn.
  */
  bool meta_size_available(void) { return meta_size_valid; };
  void meta_size(unsigned long long int val) {
    if(!meta_size_valid) { meta_size_=val; meta_size_valid=true; };
  };
  void meta_size_force(unsigned long long int val) {
    meta_size_=val; meta_size_valid=true;
  };
  unsigned long long int meta_size(void) const {
    if(meta_size_valid) return meta_size_; return 0;
  };
  bool meta_checksum_available(void) { return meta_checksum_valid; };
  void meta_checksum(unsigned long long int val) {
    if(!meta_checksum_valid) { meta_checksum_=val; meta_checksum_valid=true; };
  };
  void meta_checksum_force(unsigned long long int val) {
    meta_checksum_=val; meta_checksum_valid=true;
  };
  unsigned long long int meta_checksum(void) const {
    if(meta_checksum_valid) return meta_checksum_; return 0;
  };
  bool meta_created_available(void) { return meta_created_valid; };
  void meta_created(time_t val) {
    if(!meta_created_valid) { meta_created_=val; meta_created_valid=true; };
  };
  void meta_created_force(time_t val) {
    meta_created_=val; meta_created_valid=true;
  };
  time_t meta_created(void) const {
    if(meta_created_valid) return meta_created_; return 0;
  };
  bool meta_validtill_available(void) { return meta_validtill_valid; };
  void meta_validtill(time_t val) {
    if(!meta_validtill_valid) {meta_validtill_=val; meta_validtill_valid=true;};
  };
  void meta_validtill_force(time_t val) {
    meta_validtill_=val; meta_validtill_valid=true;
  };
  time_t meta_validtill(void) const {
    if(meta_validtill_valid) return meta_validtill_; return 0;
  };
  /*
    return true if url is "meta-url" (rc://, rls://, etc.)
  */
  bool meta(void) const { return is_rc || is_rls; };
  /*
    pass meta data from another object.
  */
  void meta(const DataPoint &p) {
    if(p.meta_size_valid)      meta_size(p.meta_size_);
    if(p.meta_checksum_valid)  meta_checksum(p.meta_checksum_);
    if(p.meta_created_valid)   meta_created(p.meta_created_);
    if(p.meta_validtill_valid) meta_validtill(p.meta_validtill_);
  };
  /*
    true if url is local (file://).
  */
  bool local(void) const { return is_file; };
  /*
    Map url (change it) according to table provided in maps.
    returns: true always.
  */
  bool map(const UrlMap &maps);
  bool sort(const UrlMap &maps);
  DataPoint& operator= (const DataPoint &);
  operator bool (void) const { return is_valid; };
  /*
    get current pfn (url).
  */
  const string& current_location(void) const { return location->url; };
  /*
    get meta information used to create curent pfn. For RC that is 
    location's name. For RLS that is equal to pfn.
  */
  const string& current_meta_location(void) const { return location->meta; };
  /*
    Switch to next location in list. At last location switch to first if
    number of tries does not exceed allowed one.
    Returns: true - will have new location, false - no more tries left.
  */
  bool next_location(void);
  /*
    true if have location to use or not resolved yet, false if out of tries. 
  */
  bool have_location(void);
  /*
    true if lfn was successfuly resolved to locations and number
    of locations is not 0.
  */
  bool have_locations(void);
  /*
    remove current location from list
  */
  bool remove_location(void);
  /*
    Get/set number of tries left.
  */
  int tries(void) { return tries_left; };
  void tries(int n) { tries_left=n; };
  /* url given to contructor */
  string base_url(void) const;
  /* 
    url given to condtructor after removing location names, 
    adding port number, etc.
  */
  string canonic_url(void) const;
};

ostream& operator<<(ostream& o,const DataPoint &point);

#endif
