#include "../../std.h"

#include <fstream>

#include "../../auth/auth.h"
#include "../../misc/escaped.h"
#include "../../transfer/proxy.h"
#include "../../misc/log_time.h"
#include "../../config/environment.h"

#include "../httpsd.h"
#ifdef HAVE_GSOAP
#include "../logger.h"
#endif

static int curr_connections = 0;
static int pre_connections = 2;
static int max_connections = 100;

globus_cond_t conn_cond;
globus_mutex_t conn_lock;

HTTP_Services services;

list<AuthEvaluator*> auths;

typedef struct {
  const char* name;
  service_creator      creat;
  service_configurator config;
} service_list_t;

service_list_t service_list[] = {
#ifdef HAVE_GSOAP
 { "logger", &logger_service_creator, &logger_service_configurator },
#endif
 { NULL, NULL, NULL }
};

int configure_services(const char* fname) {
  ifstream f(fname);
  if(!f.is_open()) {
    olog<<"Failed to read configuration file "<<fname<<endl;
    return -1;
  };
  for(;!f.eof();) {
    char buf[1024];
    istream_readline(f,buf,sizeof(buf));
    char* p = buf;
    for(;*p;p++) if(!isspace(*p)) break;
    char* command = p;
    if(*command == '#') continue;
    for(;*p;p++) if(isspace(*p)) break;
    int command_len = p-command;
    if(!command_len) continue;
    if((command_len==5) && (strncmp(command,"group",command_len)==0)) {
      string name("");
      int n = input_escaped_string(p,name);
      if(n==0) {
        cerr<<"Group name is missing"<<endl; f.close(); return -1;
      };
      p+=n;
      AuthEvaluator* auth = new AuthEvaluator(name.c_str());
      if(!auth) {
        cerr<<"Failed to create AuthEvaluator"<<endl; f.close(); return -1;
      };
      auths.push_back(auth);
      for(;!f.eof();) {
        istream_readline(f,buf,sizeof(buf));
        p = buf;
        for(;*p;p++) if(!isspace(*p)) break;
        command = p;
        if(*command == '#') continue;
        for(;*p;p++) if(isspace(*p)) break;
        command_len = p-command;
        if(!command_len) continue;
        if((command_len==3) && (strncmp(command,"end",command_len)==0)) break;
        auth->add(command);
      };
      olog<<"Added authorization group "<<name<<endl;
    } else if((command_len==7) && (strncmp(command,"service",command_len)==0)) {
      /* service name path */
      string name("");
      int n = input_escaped_string(p,name);
      if(n==0) {
        cerr<<"Service name is missing"<<endl; f.close(); return -1;
      };
      p+=n;
      string path("");
      n = input_escaped_string(p,path);
      if(path[0] != '/') path="/"+path;
      /* find service */
      service_list_t* s = service_list;
      for(;s->name;s++) if(strcmp(s->name,name.c_str()) == 0) break;
      if(!s->name) {
        cerr<<"Service "<<name<<" is not supported"<<endl; f.close(); return -1;
      };
      void* arg = s->config(f,path.c_str());
      services.add(path.c_str(),s->creat,arg);
      olog<<"Added service "<<s->name<<" at "<<path<<endl;
    } else {
      *p=0;
      cerr<<"Unknown command: "<<command<<endl; f.close(); return -1;
    };
  };
  f.close();
  return 0;
}

#ifdef OLD_GLOBUS_IO
globus_bool_t auth_cb(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t* context_handle) {
#else
globus_bool_t auth_cb(void* arg,globus_io_handle_t* h,globus_result_t result,char* identity,gss_ctx_id_t context_handle) {
#endif
  olog<<"Authenticating: "<<identity<<endl;
  HTTPS_Connector::identity(h,identity,h->delegated_credential);
  return GLOBUS_TRUE;
}

void listen_cb(void *callback_arg,globus_io_handle_t *handle,globus_result_t result) {
  if(result != GLOBUS_SUCCESS) {
    cerr<<"ERROR: listen failed "<<globus_object_printable_to_string(globus_error_get(result))<<endl;
    return;
  };
  HTTPS_Connector* c = new HTTPS_Connector(handle,services,auths);
  if(!(*c)) delete c;
  result=globus_io_tcp_register_listen(handle,&listen_cb,NULL); 
  if(result != GLOBUS_SUCCESS) {
    cerr<<"ERROR: Failed to listen "<<globus_object_printable_to_string(globus_error_get(result))<<endl;
  };
}

int main(int argc,char* argv[]) {
  unsigned short server_port = 8000;
  int handle;
  int handle_ssl;
  struct sockaddr_in myaddr;


  read_env_vars(true);

  string config_file = nordugrid_loc+"/etc/httpds.conf";

  int n;
  while((n=getopt(argc,argv,"hp:c:n:")) != -1) {
    switch(n) {
      case ':': { olog<<"Missing argument\n"; return 1; };
      case 'h': {
        fprintf(stdout,"httpsd [-p port_to_listen] [-c config_file] [-n maximal_connections]\n");
         return 0;
      };
      case 'p': {
        if(sscanf(optarg,"%hu",&server_port) != 1) {
          olog<<"Wrong port number\n";
          return 1;
        };
      }; break;
      case 'c': {
        config_file=optarg;
      }; break;
      case 'n': {
        if((sscanf(optarg,"%i",&max_connections) != 1) ||
           (max_connections < 0)) {
          olog<<"Wrong number of connections\n";
          return 1;
        };
      }; break;
      default: break;
    };
  };
  if(config_file.length() == 0) {
    olog<<"Missing configuration file"<<endl;
  };
  if(configure_services(config_file.c_str()) != 0) {
    exit(1);
  };

  if(globus_module_activate(GLOBUS_IO_MODULE) != GLOBUS_SUCCESS) {
    olog<<"Failed to activate Globus IO module"<<endl;
    return 1;
  };

  globus_io_handle_t s_gsi;
  globus_io_handle_t s_ssl;
  globus_io_attr_t attr;
  globus_io_secure_authorization_data_t auth;
  globus_result_t res;

  globus_cond_init(&conn_cond,GLOBUS_NULL);
  globus_mutex_init(&conn_lock,GLOBUS_NULL);

  globus_io_secure_authorization_data_initialize(&auth);
  globus_io_secure_authorization_data_set_callback(&auth,&auth_cb,NULL);
  globus_io_tcpattr_init(&attr);
  globus_io_attr_set_secure_authentication_mode(&attr,GLOBUS_IO_SECURE_AUTHENTICATION_MODE_GSSAPI,GSS_C_NO_CREDENTIAL);
  globus_io_attr_set_secure_authorization_mode(&attr,GLOBUS_IO_SECURE_AUTHORIZATION_MODE_CALLBACK,&auth);
  globus_io_attr_set_secure_channel_mode(&attr,GLOBUS_IO_SECURE_CHANNEL_MODE_GSI_WRAP);
  globus_io_attr_set_secure_protection_mode(&attr,GLOBUS_IO_SECURE_PROTECTION_MODE_PRIVATE);
  globus_io_attr_set_secure_delegation_mode(&attr,GLOBUS_IO_SECURE_DELEGATION_MODE_NONE);
  globus_io_attr_set_secure_proxy_mode(&attr,GLOBUS_IO_SECURE_PROXY_MODE_MANY);
  res=globus_io_tcp_create_listener(&server_port,-1,&attr,&s_gsi);   
  if(res != GLOBUS_SUCCESS) {
    cerr<<"Failed to create GSI listener "<<globus_object_printable_to_string(globus_error_get(res))<<endl;
    return 1; 
  };
  server_port++;
  globus_io_attr_set_secure_channel_mode(&attr,GLOBUS_IO_SECURE_CHANNEL_MODE_SSL_WRAP);
  res=globus_io_tcp_create_listener(&server_port,-1,&attr,&s_ssl);   
  if(res != GLOBUS_SUCCESS) {
    cerr<<"Failed to create SSL listener "<<globus_object_printable_to_string(globus_error_get(res))<<endl;
    return 1; 
  };
  res=globus_io_tcp_register_listen(&s_gsi,&listen_cb,NULL); 
  if(res != GLOBUS_SUCCESS) {
    cerr<<"Failed to listen "<<globus_object_printable_to_string(globus_error_get(res))<<endl;
    return 1; 
  };
  res=globus_io_tcp_register_listen(&s_ssl,&listen_cb,NULL); 
  if(res != GLOBUS_SUCCESS) {
    cerr<<"Failed to listen "<<globus_object_printable_to_string(globus_error_get(res))<<endl;
    return 1; 
  };

  globus_mutex_lock(&conn_lock);
  while(1) {
    globus_cond_wait(&conn_cond,&conn_lock);
  };
  globus_mutex_unlock(&conn_lock);

  globus_cond_destroy(&conn_cond);
  globus_mutex_destroy(&conn_lock);

  globus_module_deactivate(GLOBUS_IO_MODULE);

  return 0;
}

int stat_file(const char* fname,unsigned long long int &size) {
  struct stat64 st;
  if(stat64(fname,&st) != 0) return -1;
  if(!S_ISREG(st.st_mode)) return -1;
  size=st.st_size;
  return 0;
}

int send_file(const char* fname,HTTPS_Connector &c) {
  int h = ::open(fname,O_RDONLY);
  if(h == -1) return 1;
  char buf[1024];
  for(;;) {
    int l = read(h,buf,sizeof(buf));
    if(l == 0) break;
    if(l == -1) { close(h); return 1; };
    if(c.write(buf,l)) { close(h); return 1; };
  };
  close(h);
  return 0;
}

