/* 
  Filename: users.cc
  keeps list of users
*/

#include "../std.h"
#include "../config/config.h"
#include "../run/run.h"
#include "../misc/inttostring.h"
#include "../misc/log_time.h"
#include "states.h"
#include "users.h"


JobUser::JobUser(void) {
  control_dir=""; unix_name=""; home=""; cache_dir=""; cache_link_dir="";
  uid=0; gid=0; valid=false; jobs=NULL; cache_max=0; cache_min=0;
  keep_finished=DEFAULT_KEEP_FINISHED;
};

void JobUser::SetLRMS(const string &lrms_name,const string &queue_name) {
  default_lrms=lrms_name;
  default_queue=queue_name;
}

void JobUser::SetControlDir(const string &dir) {
  if(dir.length() == 0) {
    control_dir=home + "/.jobstatus";
  }
  else { control_dir=dir; };
  mkdir(control_dir.c_str(),S_IRWXU);
  mkdir((control_dir+"/logs").c_str(),S_IRWXU);
}

void JobUser::SetSessionRoot(const string &dir) {
  if(dir.length() == 0) { session_root=home + "/.jobs"; }
  else { session_root=dir; };
  mkdir(session_root.c_str(),S_IRWXU);
}

void JobUser::SetCacheDir(const string &dir,const string &data_dir,bool priv) {
  SetCacheDir(dir,data_dir,"",priv);
}

void JobUser::SetCacheSize(long long int cache_max_,long long int cache_min_) {
  if((cache_min_==0) || (cache_max_==0)) cache_min_=cache_max_;
  cache_max=cache_max_;
  cache_min=cache_min_;
}

void JobUser::SetCacheDir(const string &dir,const string &data_dir,const string &link_dir,bool priv) {
  cache_dir=dir;
  if(data_dir == "") { cache_data_dir=dir; } else { cache_data_dir=data_dir; };
  cache_link_dir=link_dir;
  private_cache=priv;
}

JobUser::JobUser(uid_t uid_) {
  struct passwd pw_;
  struct passwd *pw;
  char buf[BUFSIZ];
  uid=uid_;
  valid=false;
  /* resolve name */
  if(uid_ == 0) {
    unix_name="";  
    gid=0;
    home="/tmp";
    valid=true;
  }
  else {
    getpwuid_r(uid_,&pw_,buf,BUFSIZ,&pw);
    if(pw != NULL) {
      unix_name=pw->pw_name;  
      gid=pw->pw_gid;
      home=pw->pw_dir;
      valid=true;
    };
  };
  jobs=NULL;
  SetControlDir("");
  SetSessionRoot("");
  SetLRMS("","");
  SetCacheDir("","");
  SetCacheSize(0);
  keep_finished=DEFAULT_KEEP_FINISHED;
}

JobUser::JobUser(const string &u_name) {
  struct passwd pw_;
  struct passwd *pw;
  char buf[BUFSIZ];
  unix_name=u_name;  
  valid=false;
  /* resolve name */
  if(u_name.length() == 0) {
    uid=0;  
    gid=0;
    home="/tmp";
    valid=true;
  }
  else {
    getpwnam_r(u_name.c_str(),&pw_,buf,BUFSIZ,&pw);
    if(pw != NULL) {
      uid=pw->pw_uid;
      gid=pw->pw_gid;
      home=pw->pw_dir;
      valid=true;
    };
  };
  SetControlDir("");
  SetSessionRoot("");
  SetLRMS("","");
  SetCacheDir("","");
  SetCacheSize(0);
  jobs=NULL;
  keep_finished=DEFAULT_KEEP_FINISHED;
};

JobUser::JobUser(const JobUser &user) {
  uid=user.uid; gid=user.gid;
  unix_name=user.unix_name;  
  control_dir=user.control_dir;
  home=user.home;
  jobs=user.jobs;
  session_root=user.session_root;
  default_lrms=user.default_lrms;
  default_queue=user.default_queue;
  valid=user.valid;
  keep_finished=user.keep_finished;
  cache_dir=user.cache_dir;
  private_cache=user.private_cache;
  cache_max=user.cache_max;
  cache_min=user.cache_min;
};

JobUser::~JobUser(void) { 
};

JobUsers::JobUsers(void) {
};

JobUsers::~JobUsers(void) {
};

JobUsers::iterator JobUsers::AddUser(const string &unix_name,const string &control_dir, const string &session_root) {
  JobUser user(unix_name);
  user.SetControlDir(control_dir);
  user.SetSessionRoot(session_root);
  if(user.is_valid()) { return users.insert(users.end(),user); };
  return users.end();
};

string JobUsers::ControlDir(iterator user) {
  if(user == users.end()) return string("");
  return (*user).ControlDir();
};

JobUsers::iterator JobUsers::find(const string user) {
  iterator i;
  for(i=users.begin();i!=users.end();++i) {
    if((*i) == user) break;
  };
  return i;
}

string JobUsers::ControlDir(const string user) {
  for(iterator i=users.begin();i!=users.end();++i) {
    if((*i) == user) return (*i).ControlDir();
  };
  return string("");
}

/* change effective user - real switch is done only if running as root */
bool JobUser::SwitchUser(bool su) const {
  string uid_s = inttostring(uid);
  if(setenv("USER_ID",uid_s.c_str(),1) != 0) if(!su) return false;
  if(setenv("USER_NAME",unix_name.c_str(),1) != 0) if(!su) return false;
  /* set proper umask */
  umask(0177);
  if(!su) return true;
  uid_t cuid;
  if((cuid=getuid()) != 0) {
    if(cuid != uid) return false;
  };
  setgid(gid); /* this is not an error if group failed, not a big deal */
  if(uid != 0) if(setuid(uid) != 0) return false;
  return true;
}

bool JobUser::run_helpers(void) {
  if(unix_name.length() == 0) { /* special users can not run helpers */
    return true;  
  };
  bool started = true;
  for(list<JobUserHelper>::iterator i=helpers.begin();i!=helpers.end();++i) {
    started &= i->run(*this);
  };
  return started;
}

bool JobUsers::run_helpers(void) {
  for(iterator i=users.begin();i!=users.end();++i) {
    i->run_helpers();
  };
  return true;
}

JobUserHelper::JobUserHelper(const string &cmd) {
  command=cmd;
  proc=NULL;
}
  
JobUserHelper::~JobUserHelper(void) {
    if(proc != NULL) {
      if(proc->get_exit_code() == -1) proc->kill();
      RunParallel::release(proc);
      proc=NULL;
    };
}

bool JobUserHelper::run(JobUser &user) {
    if(proc != NULL) {
      if(proc->get_exit_code() == -1) {
        return true; /* it is already/still running */
      };
      RunParallel::release(proc);
      proc=NULL;
    };
    /* start/restart */
    if(command.length() == 0) return true;  /* has anything to run ? */
    char* args[100]; /* up to 98 arguments should be enough */
    string args_s = command;
    string arg_s;
    int n;
    for(n=0;n<99;n++) {
      arg_s=config_next_arg(args_s);
      if(arg_s.length() == 0) break;
      args[n]=strdup(arg_s.c_str());
    };
    args[n]=NULL;
    olog<<"Starting helper process ("<<user.UnixName()<<"): " <<args[0]<<endl;
    string helper_id="helper."+user.UnixName();
    bool started=RunParallel::run(user,helper_id.c_str(),args,&proc);
    for(n=0;n<99;n++) {
      if(args[n] == NULL) break;
      free(args[n]);
    };
    if(started) return true;
    olog<<"Helper process start failed: "<<command<<endl;
    /* start failed */
    /* doing nothing - maybe in the future */
    return false;
}
